package geo;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.UnsupportedEncodingException;
import java.nio.charset.Charset;

import org.mozilla.intl.chardet.nsDetector;
import org.mozilla.intl.chardet.nsICharsetDetectionObserver;
import org.mozilla.intl.chardet.nsPSMDetector;


/** Detect the character set used in a raw byte stream.
 * Tries to 1)find an html meta-tag describing charset used
 * or 2)use mozilla-chardet.
 */
public class CharsetDetector {
	

	private static String DEFAULT_CHARSET = "iso-8859-7";

	//Parsing - detector related
	private static ParsingDetector pd=null;
	private static String getCharsetParsing(InputStream bin){
		try {
			InputStreamReader in = new InputStreamReader(bin,"US-ASCII");
			
			if (pd == null) {
				pd = new ParsingDetector(in);
			} else {
				pd.yyreset(in);
			}
			
			String result = pd.getCharset();
			
			bin.reset();
						
			if (result!=null){
				if(Charset.isSupported(result)){
					return result;
				} else {
					return DEFAULT_CHARSET;
				}
			}
			
		} catch (IOException e){
			//"Impossible IOException (it cannot occur).");
			throw new RuntimeException(e);
		}
		
		return null;
	}
	
	//end parsing-detector related
	

	//Mozilla related

	private static String mozillaCharset;
	
	public static void setMozillaCharset(String charset){
		mozillaCharset = charset;
	}
	
	private static String getCharsetMozilla(byte[] input, InputStream bin){

		nsDetector det = new nsDetector(nsPSMDetector.ALL) ;
		
		mozillaCharset = null;
		
		// Set an observer...
		// The Notify() will be called when a matching charset is found.

		det.Init(new nsICharsetDetectionObserver() {
			public void Notify(String charset) {
			    CharsetDetector.setMozillaCharset(charset);
			}
	    	});
		
		boolean isAscii = det.isAscii(input,input.length);
		
		if (!isAscii) {
 		    det.DoIt(input,input.length, false);
		}

		det.DataEnd();
		
		if (!isAscii){
			if (mozillaCharset!=null){
				return correctCharsetOrDefault(mozillaCharset);
			}
			
		    String prob[] = det.getProbableCharsets() ;
		    //return first one that is greek iso, greek windows or utf8
		    for (int i = 0; i < prob.length; i++) {
				String charset = correctCharset(prob[i]);
				if (charset!= null){
					return charset;
				}				
			}
		    
		    //we failed
			return null;
		}
		
		//because it's ascii return iso-8859-7
		
		return "iso-8859-7";
	}
	
	//charset correction related:
	//if we find a charset which is not among
	//the 3 greek ones, convert it to 
	//the nearest aproximate greek one
	//e.g. windows-1252 becomes windows-1253

	//Charset correction currently used only in mozilla
	
	private static Charset isoCharset = Charset.forName("iso-8859-7");
	
	private static String correctCharset(String charset){
		//try to correct charset
		
		charset = charset.toLowerCase();
		if (charset.startsWith("windows")
					|| charset.startsWith("cp")){
			return "windows-1253";
		} else if (charset.startsWith("utf")
				||charset.startsWith("unicode")){
			if (Charset.isSupported(charset)){
				return charset;
			} else {
				//default to utf-8
				return "utf-8";
			}
		} else if (charset.startsWith("iso")
				||charset.startsWith("greek")
				||(Charset.isSupported(charset)
						&& Charset.forName(charset).equals(isoCharset))){
			return "iso-8859-7";
		} else {
			return null;
		}
	}
	
	private static String correctCharsetOrDefault(String charset){
		String result = correctCharset(charset);
		if (result!=null){
			return result;
		} else{
			return DEFAULT_CHARSET;
		}
	}
	//end mozilla related
	
	/**Convenience method to return a Reader from a charset found*/
	private static Reader charset2Reader(InputStream in, String charset,boolean detected){
		try {
			//"Detection failed. Defaulting to ") + charset);
			return new BufferedReader(new InputStreamReader(in,charset));
		} catch (UnsupportedEncodingException e){
			//charset + " not supported...");
			throw new RuntimeException(e);
		}
	}
	
	/**Return a Reader, using the proper charset, for the input given*/
	public static Reader getReader(byte[] input) {
		ByteArrayInputStream bin = new ByteArrayInputStream(input);
		String result;
		
		//Try with parsing detector
		if ((result = getCharsetParsing(bin))!=null){
			return charset2Reader(bin,result,(result!=DEFAULT_CHARSET));
			//if result == DEFAULT_CHARSET, then we failed & had to pick the default one
		}
		
		bin.reset();
		
		//Uh-oh... Try with mozilla detector
		if ((result = getCharsetMozilla(input, bin)) !=null){
			return charset2Reader(bin,result,(result!=DEFAULT_CHARSET));
		}
		
		//no need to bin.reset();
		//as mozilla doesn't actually use the reader, but the underlying
		//bytes
		
	    //we failed, default to greek iso
		return charset2Reader(bin,DEFAULT_CHARSET,false);
		
	}
	
	/**Return the charset used in the input given*/
	public static String getCharset(byte[] input){
		ByteArrayInputStream bin = new ByteArrayInputStream(input);
		String result;
		
		//Try with parsing detector
		if ((result = getCharsetParsing(bin))!=null){
			return result;
		}
		
		bin.reset();
		
		//Uh-oh... Try with mozilla detector
		if ((result = getCharsetMozilla(input, bin)) !=null){
			return result;
		}
		
		//no need to bin.reset();
		//as mozilla doesn't actually use the reader, but the underlying
		//bytes
		
	    //we failed, default to greek iso
		return DEFAULT_CHARSET;
	}
	
}
