package geo;
import java.io.IOException;
import java.io.Reader;
import java.util.HashMap;
import java.util.regex.Pattern;

import org.ccil.cowan.tagsoup.Parser;
import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.InputSource;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;

/** Parses a single html page.
 * Used through its static parse method.
 * 
 * @author Alberto Angel
 */
public class HTMLParser implements ContentHandler{

	/** A regex representing whitespace found on html pages*/
	private static Pattern whiteSpace = Pattern.compile("[ \\t\\n\\x0B\\f\\r\\x0D\\x0A\\xA0]+");
	
	private StringBuilder buffer = new StringBuilder();
	
	/** How deep are we into tags whose text shouldn't be output
	 * I.e. if we're after &lt;script> &lt;script> it should be 2*/
	private int depthOfTextIgnorableTags=0;
	
	
	/** Create a new html parser for a page.
	 * @param url Url of said page
	 */
	public HTMLParser(){
	}
	
 	/**
	 *  Delegate parsing of the page, return a
	 *  reader to the parsed page (i.e. tags stripped,
	 *  newlines inserted where appropriate, only salient
	 *  text included, whitespace trimmed etc)
	 *  
	 * @param url The url naming this page
	 * @param in Where to read the page from
	 */
	public static String getString(Reader in) throws IOException, SAXException {
		Parser tagSoupParser= new Parser();
		HTMLParser htmlParser=new HTMLParser();
		tagSoupParser.setContentHandler(htmlParser);
		InputSource is = new InputSource(in);
		
		tagSoupParser.parse(is);
		
		return htmlParser.getParsedText();
	}
		
	/**Return parsed contents and flush internal buffer*/
	public String getParsedText(){
		String result = buffer.toString();
		buffer.delete(0,buffer.length());
		return result;		
	}
	

	/**Contains the html tags which have ignorable text.
	 * Their value is immaterial.
	 * @use if(ignorableTextTags.containsKey(what.toLowerCase()))*/
	private static HashMap<String,Boolean> ignorableTextTags = new HashMap<String,Boolean>();
	static {
		//TODO: add more tags
		ignorableTextTags.put("script",null);
		ignorableTextTags.put("style",null);
	}
	
	/**Contains the html tags which don't cut the text flow.
	 * Their value is immaterial.
	 * @use if(noCutTextFlow.containsKey(what.toLowerCase()))*/
	private static HashMap<String,Boolean> noCutTextFlow = new HashMap<String,Boolean>();
	static {
		//TODO: add more tags
		noCutTextFlow.put("b",null);
		noCutTextFlow.put("i",null);
		noCutTextFlow.put("u",null);
		noCutTextFlow.put("td",null);
	}
	
	/**Contains the attribute name that might hold
	 * links in given tag, if any.
	 * @use linkType = whatLinks.get(what.toLowerCase())
	 */
	private static HashMap<String,String> whatLinks = new HashMap<String,String>();
	static{
		//TODO: add more tags
		whatLinks.put("fname","src");
		whatLinks.put("iframe","src");
		whatLinks.put("a","href");
		whatLinks.put("base","href");
		whatLinks.put("area","href");
		whatLinks.put("link","href"); //???TODO do we want this type?

	}

	/*SAX Content handler methods follow*/
	
	public void characters(char[] ch, int start, int length)
					throws SAXException{
		
		if (depthOfTextIgnorableTags <= 0) {
			
			//split tokens by hand, and feed them one-by-one
			String[] tokens = whiteSpace.split(new String(ch,start,length));
			for (int i = 0 ; i < tokens.length ; i++) {
				//assert that tokens contain no extraneous whitespace
				assert( tokens[i].equals(tokens[i].trim()));
				
				if(!tokens[i].equals("")){
					buffer.append(tokens[i]);
					buffer.append(' ');
				}
			}
		}
		
	}
	
	public void startElement(String namespaceURI,
            String localName,
            String qName,
            Attributes atts)
     throws SAXException{
		
		//TODO: might handle some tags specially, like e.g. address!
		
		String name = (localName.equals("")?qName:localName);


		if ((!noCutTextFlow.containsKey(name))&&(buffer.length()>0)&&(buffer.charAt(buffer.length()-1)!='\n')){
			buffer.append('\n');
		}

		
		
		if (ignorableTextTags.containsKey(name)){
			depthOfTextIgnorableTags++;
			return;
		}
		
		String linkAttName = whatLinks.get(name.toLowerCase());
			//the link attribute name we're looking for
		
		if ((linkAttName!=null) && (atts != null)) {

			String href="";
			
            for (int i = 0; i < atts.getLength(); i++) {
            	String atName = (atts.getLocalName(i).equals("")?atts.getQName(i):atts.getLocalName(i));
            	
            	if(atName.equalsIgnoreCase(linkAttName)) {
            		href = atts.getValue(i);
            		break; //all done checking attributes
            	}
            }
            
            if (!href.equals("")) {
            	
            	if (name.equalsIgnoreCase("base")){
                	//TODO: handle base link
            	} else {
	            	//TODO: do something with link mentioned in href.
                    //Note: take care not to get images etc, maito's, javascripts etc
            		//Link is: new URL(baseUrl,href)
            	}
            }
        }
		
		
	}
	
	public void endElement(String namespaceURI,
            String localName,
            String qName)
     throws SAXException{

		String name = (localName.equals("")?qName:localName);
		if (ignorableTextTags.containsKey(name)){
			depthOfTextIgnorableTags--;
		}
		
		if ((!noCutTextFlow.containsKey(name))&&(buffer.length()>0)&&(buffer.charAt(buffer.length()-1)!='\n')){
			buffer.append('\n');
		}
	
	}
	
	public void startDocument() throws SAXException {}
	public void endDocument() throws SAXException {}
    
	/*Non-implemented SAX ContentHandler methods*/
	
    public void ignorableWhitespace(char[] ch, int start, int length) throws SAXException{}
    public void processingInstruction(String target, String data) throws SAXException{}
	public void setDocumentLocator(Locator arg0) {}
	public void skippedEntity(String arg0) throws SAXException {}
    public void startPrefixMapping (String prefix, String uri) throws SAXException {}
    public void endPrefixMapping (String prefix) throws SAXException{}
}
